@@ -63,7 +63,7 @@
  ****************************************************************************/
 
 /*
- * <$Id: ne_io.c,v 3.20 2000/02/27 20:49:53 dillema Exp $>
+ * <$Id: ne_io.c,v 3.26 2001/03/13 11:16:50 dillema Exp $>
  */
 
 #include "tot.h"
@@ -77,10 +77,9 @@ COPYRIGHT2(
 COPYRIGHT3("Written by: F. W. Dillema. feico@pasta.cs.uit.no.\n
 based on code by: Yusuke DOI, Keio Univ. Murai Lab.");
 
-CVSID("$Id: ne_io.c,v 3.20 2000/02/27 20:49:53 dillema Exp $");
+CVSID("$Id: ne_io.c,v 3.26 2001/03/13 11:16:50 dillema Exp $");
 
-struct sockaddr *net_parse_alloc_addr (char *caddr, int port, int *sa_len_ret)
-{
+struct sockaddr *net_parse_alloc_addr (char *caddr, int port, int *sa_len_ret) {
 
 	const char *fn = "net_parse_alloc_addr()";
 	struct sockaddr *sa_p = NULL;
@@ -143,9 +142,7 @@ error:
 }
 
 int net_sockaddr_cmp (struct sockaddr * sa_p_1, int sa_len_1,
-		          struct sockaddr * sa_p_2, int sa_len_2,
-		          int ignore_flag)
-{
+		          struct sockaddr * sa_p_2, int sa_len_2, int ignore_flag) {
 #ifdef USE_INET4
 	struct sockaddr_in *sin_p_1;
 	struct sockaddr_in *sin_p_2;
@@ -192,8 +189,7 @@ int net_sockaddr_cmp (struct sockaddr * sa_p_1, int sa_len_1,
 	return -1;		/* not same */
 }
 
-int net_ifc_cmp (u_char * buf1, u_char * buf2, int buf_len)
-{
+int net_ifc_cmp (u_char * buf1, u_char * buf2, int buf_len) {
 	struct ifreq *ifr_p1;
 	struct ifreq *ifr_p2;
 	u_char *cp1;
@@ -206,8 +202,9 @@ int net_ifc_cmp (u_char * buf1, u_char * buf2, int buf_len)
 		ifr_p1 = (struct ifreq *) cp1;
 		ifr_p2 = (struct ifreq *) cp2;
 
-		DX (syslog (LOG_DEBUG, "net_ifc_cmp(): if1 %s, if2 %s\n", ifr_p1->ifr_name,
-			    ifr_p2->ifr_name));
+		if (T.debug > 3) {
+			DX (syslog (LOG_DEBUG, "net_ifc_cmp(): if1 %s, if2 %s\n", ifr_p1->ifr_name, ifr_p2->ifr_name));
+		}
 
 		/* comparison only AF_INET and AF_INET6 */
 		if (ifr_p1->ifr_addr.sa_family != ifr_p2->ifr_addr.sa_family)
@@ -241,25 +238,27 @@ int net_ifc_cmp (u_char * buf1, u_char * buf2, int buf_len)
 	return 0;
 }
 
-int net_addr_len (int af)
-{
+int net_addr_len (int af) {
 #ifdef USE_INET6
 	if (af == AF_INET6) {
-		DX (syslog (LOG_DEBUG, "%s: addrlen: %d\n", "net_addr_len()", IPV6_ADDR_LEN));
+		if (T.debug > 3) {
+			DX (syslog (LOG_DEBUG, "%s: addrlen: %d\n", "net_addr_len()", IPV6_ADDR_LEN));
+		}
 		return IPV6_ADDR_LEN;
 	}
 #endif
 #ifdef USE_INET6
 	if (af == AF_INET) {
-		DX (syslog (LOG_DEBUG, "%s: addrlen: %d\n", "net_addr_len()", IPV4_ADDR_LEN));
+		if (T.debug > 3) {
+			DX (syslog (LOG_DEBUG, "%s: addrlen: %d\n", "net_addr_len()", IPV4_ADDR_LEN));
+		}
 		return IPV4_ADDR_LEN;
 	}
 #endif
 	return -1;
 }
 
-struct sockaddr *net_sockaddr_alloc (int af, u_char * addr, int addr_len, int port)
-{
+struct sockaddr *net_sockaddr_alloc (int af, u_char * addr, int addr_len, int port) {
 	struct sockaddr *sa;
 #ifdef USE_INET4
 	struct sockaddr_in *sin;
@@ -329,8 +328,7 @@ static NI_List *NI_tail = NULL;
 static NI_List *NI_wildcard = NULL;	/* linked element of wildcard socket */
 static NI_List *NI_wildcard6 = NULL;	/* wildcard 6 */
 
-NI_List *ni_list_alloc (struct sockaddr * sa_p, int sock)
-{
+NI_List *ni_list_alloc (struct sockaddr * sa_p, int sock) {
 	const char *fn = "ni_list_alloc()";
 	NI_List *ni_tmp;
 
@@ -357,21 +355,20 @@ NI_List *ni_list_alloc (struct sockaddr * sa_p, int sock)
 	return ni_tmp;
 }
 
-void ni_free (NI_List * ni_p, int close_flag)
-{
+void ni_free (NI_List * ni_p, int close_flag) {
 	if (ni_p->sa_p)
 		free (ni_p->sa_p);
 
 	if (ni_p->sock >= 0 && close_flag) {
-		DX (syslog (LOG_DEBUG, "ni_free(): socket close(fd = %d)\n",
-			    ni_p->sock));
+		if (T.debug > 3) {
+			DX (syslog (LOG_DEBUG, "ni_free(): socket close(fd = %d)\n", ni_p->sock));
+		}
 		close (ni_p->sock);
 	}
 	free (ni_p);
 }
 
-void ni_list_destroy (NI_List * list_head)
-{
+void ni_list_destroy (NI_List * list_head) {
 	NI_List *ni_tmp;
 
 	while (list_head) {
@@ -389,8 +386,7 @@ void ni_list_destroy (NI_List * list_head)
 	return;
 }
 
-void ni_list_print (int fd)
-{
+void ni_list_print (int fd) {
 	char buf[512];
 	NI_List *ni_tmp;
 #ifdef USE_INET4
@@ -437,22 +433,21 @@ void ni_list_print (int fd)
 }
 
 
-NI_List *ni_find_by_sock (int sock_id)
-{
+NI_List *ni_find_by_sock (int sock_id) {
 	NI_List *ni_tmp;
 
 	NI_tail->sock = sock_id;
 	for (ni_tmp = NI_head; ni_tmp->sock != sock_id; ni_tmp = ni_tmp->next) {
 		;		/* just skip it */
 	}
-	if (ni_tmp == NI_tail) {
+
+	if (ni_tmp == NI_tail)
 		return NULL;
-	}
+
 	return ni_tmp;
 }
 
-NI_List *ni_find_by_addr (struct sockaddr * sa)
-{
+NI_List *ni_find_by_addr (struct sockaddr * sa) {
 	NI_List *ni_tmp;
 #ifdef USE_INET4
 	struct sockaddr_in *sin_p;
@@ -500,19 +495,21 @@ NI_List *ni_find_by_addr (struct sockaddr * sa)
 
 	if (ni_tmp == NI_tail)
 		return NULL;
+
 	return ni_tmp;
 }
 
-void ni_fds_set (fd_set * fds, int *max_fd)
-{
+void ni_fds_set (fd_set * fds, int *max_fd) {
 	NI_List *ni_tmp;
 
-	if (NI_head == NULL) {
+	if (NI_head == NULL)
 		return;
-	}
+
 	for (ni_tmp = NI_head; ni_tmp->next != NULL; ni_tmp = ni_tmp->next) {
 		if (ni_tmp->sa_p != NULL && ni_tmp->sock >= 0) {
-			DX (syslog (LOG_DEBUG, "ni_fds_set(): FD_SET(%d)", ni_tmp->sock));
+			if (T.debug > 3) {
+				DX (syslog (LOG_DEBUG, "ni_fds_set(): FD_SET(%d)", ni_tmp->sock));
+			}
 			FD_SET (ni_tmp->sock, fds);
 			if (max_fd)
 				*max_fd = MAXNUM (*max_fd, ni_tmp->sock);
@@ -523,8 +520,7 @@ void ni_fds_set (fd_set * fds, int *max_fd)
 	return;
 }
 
-NI_List *ni_fds_isset (fd_set * fds)
-{
+NI_List *ni_fds_isset (fd_set * fds) {
 	NI_List *ni_tmp;
 
 	if (!NI_head)
@@ -540,14 +536,16 @@ NI_List *ni_fds_isset (fd_set * fds)
 	return NULL;
 }
 
-int ni_set_wildsock (void)
-{
+int ni_set_wildsock (void) {
         DX(const char *fn = "ni_set_wildsock()";)
 #ifdef USE_INET4
 	struct sockaddr_in *sin_p;
 #endif
 #ifdef USE_INET6
 	struct sockaddr_in6 *sin6_p;
+#ifdef IPV6_USE_MIN_MTU
+	const int on = 1;
+#endif
 #endif
 
 #ifdef USE_INET4
@@ -578,7 +576,11 @@ int ni_set_wildsock (void)
 			free (NI_wildcard);
 			return -1;
 		}
-		DX (syslog (LOG_DEBUG, "%s: socket(IPV4) open(wildcard fd = %d)\n", fn, NI_wildcard->sock));
+
+		if (T.debug > 3) {
+			DX (syslog (LOG_DEBUG, "%s: socket(IPV4) open(wildcard fd = %d)\n", fn, NI_wildcard->sock));
+		}
+
 		/* concat to the list */
 		NI_wildcard->next = NI_head;
 		NI_head = NI_wildcard;
@@ -616,18 +618,27 @@ int ni_set_wildsock (void)
 			free (NI_wildcard6);
 			return -1;
 		}
-		DX (syslog (LOG_DEBUG, "%s: socket(IPv6) open(wildcard fd = %d)\n", fn, NI_wildcard6->sock));
+
+		if (T.debug > 3) {
+			DX (syslog (LOG_DEBUG, "%s: socket(IPv6) open(wildcard fd = %d)\n", fn, NI_wildcard6->sock));
+		}
+
 		/* concat to the list */
 		NI_wildcard6->next = NI_head;
 		NI_head = NI_wildcard6;
+
+#ifdef IPV6_USE_MIN_MTU
+		/* ignore error */
+		(void)setsockopt(NI_wildcard6->sock, IPPROTO_IPV6,
+			IPV6_USE_MIN_MTU, &on, sizeof(on));
+#endif
 	}
 #endif
 
 	return 0;
 }
 
-NI_List *ni_default (int af)
-{
+NI_List *ni_default (int af) {
 	switch (af) {
 #ifdef USE_INET4
 		case AF_INET:
@@ -643,16 +654,19 @@ NI_List *ni_default (int af)
 
 }
 
-int ni_default_sock (int af)
-{
+int ni_default_sock (int af) {
         DX(const char *fn = "ni_default_sock()";)
-	DX (syslog (LOG_DEBUG, "%s: start\n", fn));
+	if (T.debug > 3) {
+		DX (syslog (LOG_DEBUG, "%s: start\n", fn));
+	}
 
 	switch (af) {
 #ifdef USE_INET4
 	case AF_INET:
 		if (NI_wildcard) {
-			DX (syslog (LOG_DEBUG, "%s: return AF_INET.default = %d.\n", fn, NI_wildcard->sock));
+			if (T.debug > 3) {
+				DX (syslog (LOG_DEBUG, "%s: return AF_INET.default = %d.\n", fn, NI_wildcard->sock));
+			}
 			return NI_wildcard->sock;
 		}
 		break;
@@ -660,19 +674,22 @@ int ni_default_sock (int af)
 #ifdef USE_INET4
 	case AF_INET6:
 		if (NI_wildcard6) {
-			DX (syslog (LOG_DEBUG, "%s: return AF_INET6.default = %d.\n", fn, NI_wildcard6->sock));
+			if (T.debug > 3) {
+				DX (syslog (LOG_DEBUG, "%s: return AF_INET6.default = %d.\n", fn, NI_wildcard6->sock));
+			}
 			return NI_wildcard6->sock;
 		}
 		break;
 #endif
 	}
 
-	DX (syslog (LOG_DEBUG, "%s: return ERROR.\n", fn));
+	if (T.debug > 3) {
+		DX (syslog (LOG_DEBUG, "%s: return ERROR.\n", fn));
+	}
 	return -1;
 }
 
-NI_List *ni_copy (NI_List * ni_p)
-{
+NI_List *ni_copy (NI_List * ni_p) {
 	if (ni_p)
 		return ni_list_alloc (ni_p->sa_p, ni_p->sock);
 	else
@@ -691,8 +708,7 @@ NI_List *ni_copy (NI_List * ni_p)
 /* net_stream_read    : (not implemented) read data from stream   */
 /* net_stream_write   : (not implemented) write data to stream    */
 /*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
-int net_stream_socket (u_int16_t port, int af)
-{
+int net_stream_socket (u_int16_t port, int af) {
 	const char *fn = "net_stream_socket()";
 	int sock;
 	const int on = 1;
@@ -765,8 +781,7 @@ int net_stream_socket (u_int16_t port, int af)
 	return sock;
 }
 
-int net_stream_connect (u_int16_t af, u_char * addr, int addr_len, u_int16_t port)
-{
+int net_stream_connect (u_int16_t af, u_char * addr, int addr_len, u_int16_t port) {
 	const char *fn = "net_stream_connect()";
 	u_char buf[128];
 	const int on = 1;
@@ -804,7 +819,9 @@ int net_stream_connect (u_int16_t af, u_char * addr, int addr_len, u_int16_t por
 		sin_p = (struct sockaddr_in *) sa_p;
 		sin_p->sin_port = htons (port);
 		memcpy (&(sin_p->sin_addr), addr, addr_len);
-		DX (syslog (LOG_DEBUG, "%s: connect to %s(v4)", fn, inet_ntoa (sin_p->sin_addr)));
+		if (T.debug > 3) {
+			DX (syslog (LOG_DEBUG, "%s: connect to %s(v4)", fn, inet_ntoa (sin_p->sin_addr)));
+		}
 	} else
 #endif
 #ifdef USE_INET6
@@ -857,8 +874,7 @@ int net_stream_connect (u_int16_t af, u_char * addr, int addr_len, u_int16_t por
 /* net_mesg_send  : send message from specific interface           */
 /* net_mesg_recv  : (not implemented) recv mesg with input IF info */
 /*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
-NI_List *net_mesg_getif (int tcp_sock, int port_srv)
-{
+NI_List *net_mesg_getif (int tcp_sock, int port_srv) {
 	const char *fn = "net_mesg_getif()";
 	static int tcp_sock_old = -1;	/* old tcp socket if available */
 	static u_char *old_bufp = NULL;	/* allocated */
@@ -888,17 +904,25 @@ NI_List *net_mesg_getif (int tcp_sock, int port_srv)
 		return NULL;
 	}
 	if (old_bufp) {
-		DX (syslog (LOG_DEBUG, "%s: check old data\n", fn));
-		if (old_bufp_len == ifc.ifc_len
-		    && net_ifc_cmp (old_bufp, buf, old_bufp_len) == 0) {
-			DX (syslog (LOG_DEBUG, "%s: interface list unchanged.\n", fn));
+		if (T.debug > 3) {
+			DX (syslog (LOG_DEBUG, "%s: check old data\n", fn));
+		}
+
+		if (old_bufp_len == ifc.ifc_len && net_ifc_cmp (old_bufp, buf, old_bufp_len) == 0) {
+			if (T.debug > 3) {
+				DX (syslog (LOG_DEBUG, "%s: interface list unchanged.\n", fn));
+			}
 			return NI_head;
 		} else {
 			syslog (LOG_INFO, "%s: interface list changed.. let's see...\n", fn);
-			DX (syslog (LOG_DEBUG, "%s: old_bufp_len = %d and ifc.ifc_len = %d\n", fn, old_bufp_len, ifc.ifc_len));
+			if (T.debug > 3) {
+				DX (syslog (LOG_DEBUG, "%s: old_bufp_len = %d and ifc.ifc_len = %d\n", fn, old_bufp_len, ifc.ifc_len));
+			}
 		}
 	}
-	DX (syslog (LOG_DEBUG, "%s: going to check data. old_bufp = %p\n", fn, old_bufp));
+	if (T.debug > 3) {
+		DX (syslog (LOG_DEBUG, "%s: going to check data. old_bufp = %p\n", fn, old_bufp));
+	}
 
 	/* initialize NI_head */
 	ni_tmp = NULL;
@@ -906,7 +930,9 @@ NI_List *net_mesg_getif (int tcp_sock, int port_srv)
 		ni_list_destroy (NI_head);
 		NI_head = NI_tail = NI_wildcard = NULL;
 	}
-	if ((NI_tail = NI_head = ni_list_alloc (NULL, -1)) == NULL) {
+
+	NI_tail = NI_head = ni_list_alloc (NULL, -1);
+	if (!NI_tail ) {
 		syslog (LOG_ERR, "%s: can't initialize interface list.", fn);
 		return NULL;
 	}
@@ -923,22 +949,25 @@ NI_List *net_mesg_getif (int tcp_sock, int port_srv)
 #ifdef USE_INET6
 		case AF_INET6:
 			if (ifr_p->ifr_addr.sa_family == AF_INET6 && !T.ip6) {
-				DX(syslog (LOG_DEBUG, "%s: ignoring interface %s for IPv6\n", fn,
-					ifr_p->ifr_name));
+				if (T.debug > 3) {
+					DX(syslog (LOG_DEBUG, "%s: ignoring interface %s for IPv6\n", fn, ifr_p->ifr_name));
+				}
 				break;
 			}
 #endif
 #ifdef USE_INET4
 			if (ifr_p->ifr_addr.sa_family == AF_INET && !T.ip4) {
-				DX(syslog (LOG_DEBUG, "%s: ignoring interface %s for IPv4\n", fn,
-					ifr_p->ifr_name));
+				if (T.debug > 3) {
+					DX(syslog (LOG_DEBUG, "%s: ignoring interface %s for IPv4\n", fn, ifr_p->ifr_name));
+				}
 				break;
 			}
 #endif
-			DX(syslog (LOG_DEBUG, "%s: try to make IPv%c socket for interface %s\n", fn,
-			     ((ifr_p->ifr_addr.sa_family == AF_INET) ? '4' :
-			(ifr_p->ifr_addr.sa_family == AF_INET6) ? '6' : '?'),
-				ifr_p->ifr_name));
+			if (T.debug > 3) {
+				DX(syslog (LOG_DEBUG, "%s: try to make IPv%c socket for interface %s\n", fn,
+					     	((ifr_p->ifr_addr.sa_family == AF_INET) ? '4' :
+						(ifr_p->ifr_addr.sa_family == AF_INET6) ? '6' : '?'), ifr_p->ifr_name));
+			}
 
 			/* allocate new member of network interface list */
 			ni_tmp = ni_list_alloc (&ifr_p->ifr_addr, -1);
@@ -951,8 +980,6 @@ NI_List *net_mesg_getif (int tcp_sock, int port_srv)
 			if (net_mesg_socket (ni_tmp, port_srv) < 0) {
 				syslog (LOG_ERR, "%s: can't bind socket for UDP(ifname = %s).\n", fn,
 					ifr_p->ifr_name);
-				fprintf (stderr, "%s: can't bind socket for UDP(ifname = %s).\n", fn,
-					 ifr_p->ifr_name);
 				ni_free (ni_tmp, 1);
 			} else {
 				/* concat to the list */
@@ -987,8 +1014,7 @@ NI_List *net_mesg_getif (int tcp_sock, int port_srv)
 	return NI_head;
 }
 
-void net_mesg_getif_finish (void)
-{
+void net_mesg_getif_finish (void) {
 	if (NI_head) {
 		ni_list_destroy (NI_head);
 		NI_head = NULL;
@@ -996,8 +1022,7 @@ void net_mesg_getif_finish (void)
 }
 
 
-int net_mesg_socket (NI_List * ni_p, int port)
-{
+int net_mesg_socket (NI_List * ni_p, int port) {
 	const char *fn = "net_mesg_socket()";
 	int sock;
 #ifdef USE_INET4
@@ -1030,7 +1055,9 @@ int net_mesg_socket (NI_List * ni_p, int port)
 		syslog (LOG_ERR, "%s: socket open failed: %m", fn);
 		return -1;
 	}
-	DX (syslog (LOG_DEBUG, "%s: socket open(fd = %d)\n", fn, sock));
+	if (T.debug > 3) {
+		DX (syslog (LOG_DEBUG, "%s: socket open(fd = %d)\n", fn, sock));
+	}
 
 	if (bind (sock, ni_p->sa_p, SOCKADDR_SIZEOF (*ni_p->sa_p)) < 0) {
 		syslog (LOG_ERR, "%s: can't bind dgram socket: %m", fn);
@@ -1039,85 +1066,50 @@ int net_mesg_socket (NI_List * ni_p, int port)
 	}
 	ni_p->sock = sock;
 
+#ifdef USE_INET6
+#ifdef IPV6_USE_MIN_MTU
+	if (ni_p->sa_p->sa_family == PF_INET6) {
+		const int on = 1;
+
+		/* ignore error */
+		(void)setsockopt(NI_wildcard6->sock, IPPROTO_IPV6,
+			IPV6_USE_MIN_MTU, &on, sizeof(on));
+	}
+#endif
+#endif
+
 	return sock;
 }
 
-int net_mesg_send (NI_List * ni_send, u_char * mesg, int mesg_len,
-		       struct sockaddr * dest_p)
-{
+int net_mesg_send (NI_List * ni_send, u_char * mesg, int mesg_len, struct sockaddr *sa_p) {
 	const char *fn = "net_mesg_send()";
-	struct sockaddr *sa_p;
-#ifdef USE_INET6
-	struct sockaddr_in *sin_p;
-	struct sockaddr_in6 dest6;
-#endif
 
 	if (!ni_send)
-		ni_send = ni_default (dest_p->sa_family);
+		ni_send = ni_default (sa_p->sa_family);
 
-	if (!ni_send || ni_send->sock < 0) {
+	if (!ni_send || ni_send->sock < 0) 
 		syslog (LOG_WARNING, "%s: attempts to send message from interface which have no socket.", fn);
-		return -1;
-	}
-#ifndef NDEBUG
-	{
-#ifdef USE_INET4
-		struct sockaddr_in *sin;
-		if (dest_p->sa_family == AF_INET) {
-			sin = (struct sockaddr_in *) dest_p;
-			syslog (LOG_DEBUG, "%s: UDP-sendto(%s)", fn, inet_ntoa (sin->sin_addr));
-#endif
-#ifdef USE_INET6
-		} else if (dest_p->sa_family == AF_INET6) {
-		        struct sockaddr_in6 *sin6;
-		        char addrstr[64];
-
-			sin6 = (struct sockaddr_in6 *) dest_p;
-			inet_ntop (AF_INET6, (const void *) &(sin6->sin6_addr), addrstr, 64);
-			syslog (LOG_DEBUG, "%s: UDP-sendto(%s)\n", fn, addrstr);
-#endif
-		}
-	}
-#endif				/* NDEBUG */
 
 #ifdef USE_INET4
 	if (ni_send->sa_p->sa_family == AF_INET) {
-		if (dest_p->sa_family == AF_INET6) {
+		if (sa_p->sa_family == AF_INET)
+			return sendto (ni_send->sock, mesg, mesg_len, 0, sa_p, SOCKADDR_SIZEOF (*sa_p));
+		else
 			syslog (LOG_NOTICE, "%s: can't send to IPv6 address with IPv4 socket.\n", fn);
-			return -1;
-		}
-		sa_p = dest_p;
 	} else
 #endif
 #ifdef USE_INET6
 	if (ni_send->sa_p->sa_family == AF_INET6) {
-		if (dest_p->sa_family == AF_INET) {
-			/* data convertion needed */
-			/*
-			 * XXX #warning -- sockaddr conversion doesnot work
-			 * -- so now I have socket for both.
-			 */
-			sin_p = (struct sockaddr_in *) dest_p;
-			memset (&dest6, 0, sizeof (dest6));
-			dest6.sin6_len = sizeof (dest6);
-			dest6.sin6_family = AF_INET6;
-			memcpy (&(dest6.sin6_addr.s6_addr[12]), &(sin_p->sin_addr),
-				sizeof (u_int32_t));
-			sa_p = (struct sockaddr *) & dest6;
-		} else if (dest_p->sa_family == AF_INET6)
-			sa_p = dest_p;
-		else {
-			syslog (LOG_ERR, "%s: unknown dest address family %d\n", fn,
-				dest_p->sa_family);
-			return -1;
-		}
+		if (sa_p->sa_family == AF_INET6) 
+			return sendto (ni_send->sock, mesg, mesg_len, 0, sa_p, SOCKADDR_SIZEOF (*sa_p));
+		else
+			syslog (LOG_ERR, "%s: can't send to IPv4 address with IPv6 socket.\n", fn);
 	} else
 #endif
 	{
-		syslog (LOG_ERR, "%s: wrong I/F address family %d\n", fn,
+		syslog (LOG_ERR, "%s: unsupported address family %d\n", fn,
 			ni_send->sa_p->sa_family);
-		return -1;
 	}
 
-	return sendto (ni_send->sock, mesg, mesg_len, 0, sa_p, SOCKADDR_SIZEOF (*sa_p));
+	return -1;
 }
