@@ -63,7 +63,7 @@
  ****************************************************************************/
 
 /*
- * <$Id: ne_mesg.c,v 3.10 1999/08/17 16:52:10 feico Exp $>
+ * <$Id: ne_mesg.c,v 3.15 2001/03/13 11:14:43 dillema Exp $>
  */
 
 
@@ -78,11 +78,10 @@ COPYRIGHT2(
 COPYRIGHT3("Written by: F. W. Dillema. feico@pasta.cs.uit.no.\n
 based on code by: Yusuke DOI, Keio Univ. Murai Lab.");
 
-CVSID("$Id: ne_mesg.c,v 3.10 1999/08/17 16:52:10 feico Exp $");
+CVSID("$Id: ne_mesg.c,v 3.15 2001/03/13 11:14:43 dillema Exp $");
 
 
-u_int16_t mesg_id (void)
-{
+u_int16_t mesg_id (void) {
 	static u_int16_t id = 0;
 
 	if (!id) {
@@ -90,21 +89,23 @@ u_int16_t mesg_id (void)
 		id = random ();
 	}
 	id++;
-	DX (syslog (LOG_DEBUG, "mesg_id() = %d", id));
+
+	if (T.debug > 3) {
+		DX (syslog (LOG_DEBUG, "mesg_id() = %d", id));
+	}
 	return id;
 }
 
-int mesg_make_query (u_char * qname, u_int16_t qtype, u_int16_t qclass, u_int32_t id,
-		         int rd, u_char * buf, int buflen)
-{
-	DX (char *fn = "mesg_make_query()";
-	)
-		register u_char *ucp;
+int mesg_make_query (u_char *qname, u_int16_t qtype, u_int16_t qclass, u_int32_t id,
+		         int rd, u_char *buf, int buflen) {
+	DX (char *fn = "mesg_make_query()";)
+	u_char *ucp;
 	int i, written_len;
 	Mesg_Hdr *hdr;
 
-	DX (syslog (LOG_DEBUG, "%s: (qtype: %s, id: %d): start",
-		    fn, string_rtype (qtype), id));
+	if (T.debug > 3) {
+		DX (syslog (LOG_DEBUG, "%s: (qtype: %s, id: %d): start", fn, string_rtype (qtype), id));
+	}
 
 	hdr = (Mesg_Hdr *) buf;
 
@@ -126,7 +127,10 @@ int mesg_make_query (u_char * qname, u_int16_t qtype, u_int16_t qclass, u_int32_
 	ucp = (u_char *) (hdr + 1);
 
 	/* write qname */
-	DX (syslog (LOG_DEBUG, "%s: qname offset = %d\n", fn, ucp - buf));
+	if (T.debug > 3) {
+		DX (syslog (LOG_DEBUG, "%s: qname offset = %d\n", fn, ucp - buf));
+	}
+
 	i = mesg_dname_copy (qname, ucp, buflen - written_len);
 	if (i < 0)
 		return -1;
@@ -135,7 +139,9 @@ int mesg_make_query (u_char * qname, u_int16_t qtype, u_int16_t qclass, u_int32_
 	ucp += i;
 
 	/* write qtype / qclass */
-	DX (syslog (LOG_DEBUG, "%s: qtype/qclass offset = %d\n", fn, ucp - buf));
+	if (T.debug > 3) {
+		DX (syslog (LOG_DEBUG, "%s: qtype/qclass offset = %d\n", fn, ucp - buf));
+	}
 	written_len += sizeof (u_int16_t) * 2;
 	if (written_len > buflen)
 		return -1;
@@ -147,13 +153,13 @@ int mesg_make_query (u_char * qname, u_int16_t qtype, u_int16_t qclass, u_int32_
 }
 
 
-int mesg_dname_copy (u_char * from, u_char * to, int tolen)
-{
+int mesg_dname_copy (u_char *from, u_char *to, int tolen) {
 	int skip, written_len;
 
 	written_len = 0;
 	while (*from) {
-		written_len += (skip = (*from) + 1);
+		skip = mesg_labellen(from) + 1;
+		written_len += skip;
 		if (written_len >= tolen)
 			return -1;
 
@@ -167,8 +173,7 @@ int mesg_dname_copy (u_char * from, u_char * to, int tolen)
 	return written_len;
 }
 
-u_char *mesg_dname_redirect (u_char * label, u_char * mesg_head)
-{
+u_char *mesg_dname_redirect (u_char * label, u_char * mesg_head) {
 	u_int16_t us_tmp;
 
 	if (mesg_head && (*label & DNCMP_MASK) == DNCMP_MASK) {
@@ -181,11 +186,9 @@ u_char *mesg_dname_redirect (u_char * label, u_char * mesg_head)
 
 int mesg_write_dname (u_char * buf_head, u_char * buf_tail,
 		          u_int16_t * offset, int offset_len,
-		          u_char * dname, u_char * wp)
-{
-	DX (char *fn = "mesg_write_dname()";
-	)
-		int written_len, match_len;
+		          u_char * dname, u_char * wp) {
+	DX (char *fn = "mesg_write_dname()";)
+	int written_len, match_len;
 	u_char *rp_dname, *cp_dname;
 	u_char *rp_mesg, *cp_mesg;
 	int i, bestmatch_len = 0;
@@ -193,35 +196,65 @@ int mesg_write_dname (u_char * buf_head, u_char * buf_tail,
 	u_char *bestmatch_mesgp;
 	u_int16_t us_tmp;
 
-	DX (syslog (LOG_DEBUG, "%s: start\n", fn));
+	if (T.debug > 3) {
+		DX (syslog (LOG_DEBUG, "%s: start\n", fn));
+	}
 
 	/** check if some (part of) dname has already appeared in message */
 	rp_dname = dname;
-	FOR_DNAME (rp_dname) {
+	while(*rp_dname != '\0') {
+		int dlen = mesg_labellen(rp_dname);
+ 
 		for (i = 0; offset[i] != 0 && i < offset_len; i++) {
 			rp_mesg = buf_head + offset[i];
-			FOR_DNAME (rp_mesg) {
+			while(*rp_mesg != '\0') {
+				int rplen;
+
 				cp_dname = rp_dname;
 				/* check redirection */
 				cp_mesg = rp_mesg = mesg_dname_redirect (rp_mesg, buf_head);
+				rplen = mesg_labellen(rp_mesg);
 
 				if (cp_mesg < buf_head || buf_tail < cp_mesg) {
-					DX (syslog (LOG_DEBUG, "out of bounds(1): buf_head = 0x%p buf_tail = 0x%p cp_mesg = 0x%p\n", (void *) buf_head, (void *) buf_tail, (void *) cp_mesg));
+					if (T.debug > 3) {
+						DX (syslog (LOG_DEBUG, "out of bounds(1): buf_head = 0x%p buf_tail = 0x%p cp_mesg = 0x%p\n", 
+						            (void *) buf_head, (void *) buf_tail, (void *) cp_mesg));
+					}
 					return -1;
 				}
+
+				/* now cp_dname is not compressed. */
 				match_len = 0;
 				/* start compare */
-				while (*cp_mesg != '\0' && *cp_dname != '\0'
-				       && *cp_mesg == *cp_dname
-				       && strncasecmp (cp_mesg + 1, cp_dname + 1, *cp_mesg) == 0) {
+				while (*cp_mesg != '\0' &&
+				       *cp_mesg == *cp_dname) {
+					int mlen = mesg_labellen(cp_mesg);
+					int dlen0 = mesg_labellen(cp_dname);
+
+					if (dlen0 != mlen)
+						break;
+					if (*cp_mesg == EDNS0_ELT_BITLABEL) {
+						if (memcmp(cp_mesg + 1,
+							   cp_dname + 1,
+							   mlen) != 0)
+							break;
+					} else if (strncasecmp (cp_mesg + 1,
+								cp_dname + 1,
+								*cp_mesg) != 0) {
+						break;
+					}
+					
 					/* a label match */
-					cp_mesg += *cp_mesg + 1;
-					cp_dname += *cp_dname + 1;
+					cp_mesg += mlen + 1;
+					cp_dname += mlen + 1;
 					/* check redirection */
 					cp_mesg = mesg_dname_redirect (cp_mesg, buf_head);
 					if (cp_mesg < buf_head || buf_tail < cp_mesg) {
 						/* out of bounds */
-						DX (syslog (LOG_DEBUG, "out of bounds(2): buf_head = 0x%p buf_tail = 0x%p cp_mesg = 0x%p\n", (void *) buf_head, (void *) buf_tail, (void *) cp_mesg));
+						if (T.debug > 3) {
+							DX (syslog (LOG_DEBUG, "out of bounds(2): buf_head = 0x%p buf_tail = 0x%p cp_mesg = 0x%p\n", 
+								    (void *) buf_head, (void *) buf_tail, (void *) cp_mesg));
+						}
 						return -1;
 					}
 					match_len++;
@@ -234,6 +267,8 @@ int mesg_write_dname (u_char * buf_head, u_char * buf_tail,
 						bestmatch_len = match_len;
 					}
 				}
+
+				rp_mesg += (rplen + 1);
 			}
 		}
 
@@ -241,6 +276,8 @@ int mesg_write_dname (u_char * buf_head, u_char * buf_tail,
 			/* don't check *shorter* value */
 			break;
 		}
+
+		rp_dname += (dlen + 1);
 	}
 
 	/** register this name if not complete match */
@@ -252,16 +289,20 @@ int mesg_write_dname (u_char * buf_head, u_char * buf_tail,
 	    && ((u_int16_t) (bestmatch_mesgp - buf_head) < DNCMP_REDIRECT_LIMIT)) {
 		offset[i] = (u_int16_t) (wp - buf_head);
 		offset[i + 1] = 0;
-		DX (syslog (LOG_DEBUG, "%s: study offset[%d] = %d\n", fn, i, offset[i]));
+		if (T.debug > 3) {
+			DX (syslog (LOG_DEBUG, "%s: study offset[%d] = %d\n", fn, i, offset[i]));
+		}
 	}
 	/** write dname */
 	written_len = 0;
 	cp_dname = dname;
 	while (*cp_dname && cp_dname != bestmatch_dnamep) {
-		i = (*cp_dname) + 1;
+		i = mesg_labellen(cp_dname) + 1;
 		if (wp + i > buf_tail) {
 			/* overflow! */
-			DX (syslog (LOG_DEBUG, "%s: overflow!\n", fn));
+			if (T.debug > 3) {
+				DX (syslog (LOG_DEBUG, "%s: overflow!\n", fn));
+			}
 			return -1;
 		}
 		memcpy (wp, cp_dname, i);
@@ -273,7 +314,9 @@ int mesg_write_dname (u_char * buf_head, u_char * buf_tail,
 		/* write redirection pointer */
 		if (wp + sizeof (u_int16_t) > buf_tail) {
 			/* overflow! */
-			DX (syslog (LOG_DEBUG, "%s: overflow!\n", fn));
+			if (T.debug > 3) {
+				DX (syslog (LOG_DEBUG, "%s: overflow!\n", fn));
+			}
 			return -1;
 		}
 		us_tmp = (u_int16_t) (bestmatch_mesgp - buf_head) | DNCMP_MASK_INT16T;
@@ -285,35 +328,77 @@ int mesg_write_dname (u_char * buf_head, u_char * buf_tail,
 		written_len++;
 	}
 
-	DX (syslog (LOG_DEBUG, "%s: return (written_len = %d)\n", fn, written_len));
+	if (T.debug > 3) {
+		DX (syslog (LOG_DEBUG, "%s: return (written_len = %d)\n", fn, written_len));
+	}
 	return written_len;
 }
 
-u_char *mesg_skip_dname (u_char * dname)
+int
+mesg_labellen(const u_char *cp)
 {
-	for ( /* void */ ; *dname; dname += ((*dname) + 1))
-		if ((*dname & DNCMP_MASK) != 0) {
+	register u_int i;
+
+	i = *cp;
+	if ((i & DNCMP_MASK) == EDNS0_MASK) {
+		int bitlen;
+
+		if (i != EDNS0_ELT_BITLABEL)
+			return(-1);
+		if ((bitlen = *(cp + 1)) == 0)
+			bitlen = 256;
+		return(((bitlen + 7) / 8) + 1);
+	} else
+		return(i);
+}
+
+u_char *mesg_skip_dname (u_char * dname, u_char *end) {
+	int l;
+
+	if (dname >= end)
+		return(NULL);
+
+	while(*dname) {
+		if ((*dname & DNCMP_MASK) == DNCMP_MASK) {
 			dname += 2;	/* redirection */
 			return dname;
 		}
+		if (dname + 2 > end) /* we need at least 2 bytes */
+			return(NULL);
+
+		if ((l = mesg_labellen(dname)) < 0)
+			return(NULL);
+		dname += l + 1;
+
+		if (dname >= end)
+			return(NULL);
+	}
 	dname++;		/* push away null terminator */
 	return dname;
 }
 
-int mesg_dname_cmp (u_char * mesg_head, u_char * dname_mesg,
-		        u_char * dname)
-{
-	int len;
-
+int mesg_dname_cmp (u_char * mesg_head, u_char * dname_mesg, u_char * dname) {
 	dname_mesg = mesg_dname_redirect (dname_mesg, mesg_head);
 
 	/* compare label by label */
-	while ((len = *dname_mesg) != '\0' && (*dname == *dname_mesg)) {
-		if (strncasecmp (dname_mesg + 1, dname + 1, len) != 0)
-			return -1;
+	while (*dname_mesg != '\0' && (*dname == *dname_mesg)) {
+		int len0 = mesg_labellen(dname_mesg);
+		int len1 = mesg_labellen(dname);
 
-		dname += len + 1;
-		dname_mesg += len + 1;	/* <- plus one is for length byte
+		if (len0 != len1)
+			return(-1);
+
+		if (*dname == EDNS0_ELT_BITLABEL) {
+			if (memcmp(dname_mesg + 1, dname + 1,  len0) != 0)
+				return -1;
+		}
+		else {
+			if (strncasecmp (dname_mesg + 1, dname + 1, len0) != 0)
+				return -1;
+		}
+
+		dname += len0 + 1; /* note that len0 == len1 */
+		dname_mesg += len0 + 1;	/* <- plus one is for length byte
 					 * (see RFC1035) */
 		dname_mesg = mesg_dname_redirect (dname_mesg, mesg_head);
 	}
@@ -326,11 +411,9 @@ int mesg_dname_cmp (u_char * mesg_head, u_char * dname_mesg,
 
 int mesg_write_rrset_list (G_List * rrls, u_char * buf_head,
 			       u_char * buf_tail, u_int16_t * offset,
-			       int offset_len, u_char ** wp, u_int16_t * cnt)
-{
-	DX (char *fn = "mesg_write_rrset_list()";
-	)
-		u_char *wp_start, *wp_period;
+			       int offset_len, u_char ** wp, u_int16_t * cnt) {
+	DX (char *fn = "mesg_write_rrset_list()";)
+	u_char *wp_start, *wp_period;
 	RRset *rrsp;
 	RR *rrp;
 	int i, ret;
@@ -338,13 +421,17 @@ int mesg_write_rrset_list (G_List * rrls, u_char * buf_head,
 	u_int16_t us_tmp;
 	u_int32_t ul_tmp;
 
-	DX (syslog (LOG_DEBUG, "%s: start.\n", fn));
+	if (T.debug > 3) {
+		DX (syslog (LOG_DEBUG, "%s: start.\n", fn));
+	}
 
 	hdr = (Mesg_Hdr *) buf_head;
 	wp_start = wp_period = *wp;
 
 	for (rrls = rrls->next; rrls->list_data; rrls = rrls->next) {
-		DX (syslog (LOG_DEBUG, "%s: write a record.\n", fn));
+		if (T.debug > 3) {
+			DX (syslog (LOG_DEBUG, "%s: write a record.\n", fn));
+		}
 		for (i = 0; i < ((RRset *) (rrls->list_data))->data.d->data_cnt; i++) {
 			rrsp = (RRset *) rrls->list_data;
 			wp_period = *wp;
@@ -353,7 +440,9 @@ int mesg_write_rrset_list (G_List * rrls, u_char * buf_head,
 			ret = mesg_write_dname (buf_head, buf_tail, offset, offset_len,
 					  rrset_key_info_owner (rrsp), *wp);
 			if (ret < 0) {
-				DX (syslog (LOG_DEBUG, "%s: fail to write owner name.\n", fn));
+				if (T.debug > 3) {
+					DX (syslog (LOG_DEBUG, "%s: fail to write owner name.\n", fn));
+				}
 				*wp = wp_period;
 				return wp_period - wp_start;
 			}
@@ -361,10 +450,11 @@ int mesg_write_rrset_list (G_List * rrls, u_char * buf_head,
 
 			/* write RR field and data */
 			rrp = (RR *) (rrsp->data.p + data_offset (i, rrsp->data.p));
-			if (*wp + sizeof (u_int16_t) * 3 + sizeof (u_int32_t) + rrp->rd_len
-			    > buf_tail) {
-				DX (syslog (LOG_DEBUG, "%s: can't write Rdata; drop owners etc\n", fn));
-				DX (syslog (LOG_DEBUG, "%s: return %d\n", fn, wp_period - wp_start));
+			if (*wp + sizeof (u_int16_t) * 3 + sizeof (u_int32_t) + rrp->rd_len > buf_tail) {
+				if (T.debug > 3) {
+					DX (syslog (LOG_DEBUG, "%s: can't write Rdata; drop owners etc\n", fn));
+					DX (syslog (LOG_DEBUG, "%s: return %d\n", fn, wp_period - wp_start));
+				}
 				*wp = wp_period;
 				return wp_period - wp_start;
 			}
@@ -382,18 +472,22 @@ int mesg_write_rrset_list (G_List * rrls, u_char * buf_head,
 			/* update header */
 			us_tmp = ntohs (*cnt) + 1;
 			*cnt = htons (us_tmp);
-			DX (syslog (LOG_DEBUG, "%s: now counter = %d", fn, us_tmp));
+
+			if (T.debug > 3) {
+				DX (syslog (LOG_DEBUG, "%s: now counter = %d", fn, us_tmp));
+			}
 		}
 	}
 
-	DX (syslog (LOG_DEBUG, "%s: return %d\n", fn, *wp - wp_start));
+	if (T.debug > 3) {
+		DX (syslog (LOG_DEBUG, "%s: return %d\n", fn, *wp - wp_start));
+	}
 	return (*wp - wp_start);
 }
 
 int mesg_assemble (Mesg_Hdr * hdr, G_List * an_list, G_List * ns_list,
 		       G_List * ar_list, u_char * buf, u_int16_t buflen,
-		       int mesg_len)
-{
+		       int mesg_len) {
 
 	int written_len = 0;
 	int ret;
@@ -416,6 +510,8 @@ int mesg_assemble (Mesg_Hdr * hdr, G_List * an_list, G_List * ns_list,
 	hdr->arcnt = 0;
 
 	if (hdr->qdcnt) {
+		int qdcnt = ntohs(hdr->qdcnt);
+
 		/* register question name to compression name list */
 		/*
 		 * XXX -- ignore multiple question :p -- I think minor
@@ -424,9 +520,24 @@ int mesg_assemble (Mesg_Hdr * hdr, G_List * an_list, G_List * ns_list,
 		offset[0] = (u_int16_t) (sizeof (Mesg_Hdr));
 		offset[1] = 0;	/* <- terminator */
 
-		/* already the question written in buffer */
-		written_len = mesg_len;
-		ucp = buf + mesg_len;
+		/*
+		 * already the question written in buffer. Skip the question
+		 * section.
+		 */
+		ucp = buf + sizeof(Mesg_Hdr);
+		while(qdcnt--) {
+			/* skip QNAME */
+			ucp = mesg_skip_dname(ucp, buf + mesg_len);
+			if (ucp == NULL ||
+			    ucp + 2 * sizeof(u_int16_t) > buf + mesg_len) {
+				syslog (LOG_NOTICE,
+					"original query message overrun\n");
+				return(-1);
+			}
+			/* skip QTYPE and QCLASS */
+			ucp += (2 * sizeof(u_int16_t));
+		}
+		written_len = ucp - buf;
 	} else {
 		/* there mustnot be any question */
 		written_len = sizeof (Mesg_Hdr);
@@ -472,8 +583,7 @@ int mesg_assemble (Mesg_Hdr * hdr, G_List * an_list, G_List * ns_list,
 	return written_len;
 }
 
-void rrset_couple_free (RRset_Couple * rc)
-{
+void rrset_couple_free (RRset_Couple * rc) {
 
 	if (rc->rrs)
 		rrset_free (rc->rrs);
@@ -483,14 +593,13 @@ void rrset_couple_free (RRset_Couple * rc)
 	free (rc);
 }
 
-void rrset_couple_freev (void *rcv)
-{
+void rrset_couple_freev (void *rcv) {
 	rrset_couple_free ((RRset_Couple *) rcv);
 }
 
+/* XXX: unused? */
 int mesg_rdata_cmp (u_int16_t r_type, u_int16_t rd_len, u_char * mesg_head,
-		        u_char * rd_mesg, u_char * rd)
-{
+		        u_char * rd_mesg, u_char * rd) {
 	char *fn = "mesg_rdata_cmp()";
 	u_char *rp, *rp_mesg;
 
@@ -558,8 +667,7 @@ int mesg_rdata_cmp (u_int16_t r_type, u_int16_t rd_len, u_char * mesg_head,
 
 int mesg_extract_rr (u_char * begin, u_char * end, u_int16_t r_type,
 		         u_int16_t r_class, u_char * rrp, u_char * buf,
-		         int buflen)
-{
+		         int buflen) {
 	u_char *rp, *wp;
 	int i, written_len = 0;
 
@@ -626,6 +734,7 @@ int mesg_extract_rr (u_char * begin, u_char * end, u_int16_t r_type,
 	case RT_A:
 	case RT_HINFO:
 	case RT_AAAA:
+	case RT_A6:
 	case RT_SRV:
 	case RT_TXT:
 		/* no modification */
@@ -640,8 +749,7 @@ int mesg_extract_rr (u_char * begin, u_char * end, u_int16_t r_type,
 	return written_len;
 }
 
-u_char *mesg_read_sec (int cnt, u_char * sec, u_char * begin, u_char * end, G_List * ls)
-{
+u_char *mesg_read_sec (int cnt, u_char * sec, u_char * begin, u_char * end, G_List * ls) {
 	char *fn = "mesg_read_sec()";
 	G_List *rc_head, *gl_tmp;
 	RRset_Couple *rc_tmp;
@@ -655,19 +763,28 @@ u_char *mesg_read_sec (int cnt, u_char * sec, u_char * begin, u_char * end, G_Li
 	u_int32_t r_ttl;
 	u_char *rdp;
 
-	DX (syslog (LOG_DEBUG, "%s: start", fn));
+	if (T.debug > 3) {
+		DX (syslog (LOG_DEBUG, "%s: start", fn));
+	}
+
 	/* initialize */
 	rc_head = list_init ();
 	if (!rc_head)
 		return NULL;
 	cnt = ntohs (cnt);
 
-	DX (syslog (LOG_DEBUG, "%s: read and decompress RDATA", fn));
+	if (T.debug > 3) {
+		DX (syslog (LOG_DEBUG, "%s: read and decompress RDATA", fn));
+	}
+
 	rp = sec;
 	for (i = 0; i < cnt; i++) {
 		/* parse header */
 		rname = rp;
-		rp = mesg_skip_dname (rp);
+		if ((rp = mesg_skip_dname (rp, end)) == NULL)
+			goto bad;
+		if (rp + sizeof(u_int16_t) * 3 + sizeof(u_int32_t) > end)
+			goto bad;
 		GETSHORT (r_type, rp);
 		GETSHORT (r_class, rp);
 		GETLONG (r_ttl, rp);
@@ -675,37 +792,46 @@ u_char *mesg_read_sec (int cnt, u_char * sec, u_char * begin, u_char * end, G_Li
 		rdp = rp;
 		rp += rdlen;
 		if (rp > end) {
+		  bad:
 			/* overrun!! */
 			syslog (LOG_WARNING, "%s: message overrun", fn);
-			DX (syslog (LOG_DEBUG, "%s: i = %d, begin = 0x%p, end-begin = 0x%p,\
- rdp - begin = 0x%p, rdlen = 0x%d, rname - begin = 0x%p",
+			if (T.debug > 3) {
+				DX (syslog (LOG_DEBUG, "%s: i = %d, begin = 0x%p, end-begin = 0x%p, rdp - begin = 0x%p, rdlen = 0x%d, rname - begin = 0x%p",
 				    fn, i, begin, (void *) (end - begin), (void *) (rdp - begin),
 				    rdlen, (void *) (rname - begin)));
+			}
 			goto error;
 		}
-		DX (syslog (LOG_DEBUG, "%s: seek for matching RRset_Couple", fn));
+		if (T.debug > 3) {
+			DX (syslog (LOG_DEBUG, "%s: seek for matching RRset_Couple", fn));
+		}
 		rc_head->list_data = NULL;
 		for (gl_tmp = rc_head->next; gl_tmp->list_data; gl_tmp = gl_tmp->next) {
 			rc_tmp = (RRset_Couple *) (gl_tmp->list_data);
 			if ((rc_tmp->rrs->key.info->r_type == r_type) &&
 			    (rc_tmp->rrs->key.info->r_class == r_class) &&
 			    (mesg_dname_cmp (begin, rname, rrset_key_info_owner (rc_tmp->rrs)) == 0)) {
-				DX (syslog (LOG_DEBUG,
-					    "%s: matching record find. rrs->dname = %s / rname = %s",
-					    fn, rrset_key_info_owner (rc_tmp->rrs), rname));
+				if (T.debug > 3) {
+					DX (syslog (LOG_DEBUG, "%s: matching record find. rrs->dname = %s / rname = %s",
+					            fn, rrset_key_info_owner (rc_tmp->rrs), rname));
+				}
 				break;
 			}
 		}
 		if (!gl_tmp->list_data) {
 			/* extract domain name */
-			DX (syslog (LOG_DEBUG, "%s: extract domain name", fn));
+			if (T.debug > 3) {
+				DX (syslog (LOG_DEBUG, "%s: extract domain name", fn));
+			}
 			if (!conv_dname_decompress (buf, sizeof (buf), rname,
 						  begin, end, &dname_len)) {
 				syslog (LOG_WARNING, "%s: dname decompression error", fn);
 				goto error;
 			}
 			/* create new one */
-			DX (syslog (LOG_DEBUG, "%s: create new entry.", fn));
+			if (T.debug > 3) {
+				DX (syslog (LOG_DEBUG, "%s: create new entry.", fn));
+			}
 
 			if (((rc_tmp = malloc (sizeof (RRset_Couple))) == NULL)
 			    || ((rc_tmp->rrl = rr_list_alloc ()) == NULL)
@@ -721,10 +847,16 @@ u_char *mesg_read_sec (int cnt, u_char * sec, u_char * begin, u_char * end, G_Li
 				rc_tmp = NULL;
 				goto error;
 			}
-		} else
-			DX (syslog (LOG_DEBUG, "%s: re-use old entry (RRset_Couple).", fn));
+		} else {
+			if (T.debug > 3) {
+				DX (syslog (LOG_DEBUG, "%s: re-use old entry (RRset_Couple).", fn));
+			}
+		}
+
+		if (T.debug > 3) {
+			DX (syslog (LOG_DEBUG, "%s:  extract RR.", fn));
+		}
 
-		DX (syslog (LOG_DEBUG, "%s:  extract RR.", fn));
 		if (rdlen) {
 			ret = mesg_extract_rr (begin, end, r_type, r_class, rdp, buf, sizeof (buf));
 			if (ret < 0) {
@@ -738,12 +870,16 @@ u_char *mesg_read_sec (int cnt, u_char * sec, u_char * begin, u_char * end, G_Li
 				rdlen_ex = ret;
 			}
 		} else {
-			DX (syslog (LOG_DEBUG, "%s:  rdlen == 0 -- update?", fn));
+			if (T.debug > 3) {
+				DX (syslog (LOG_DEBUG, "%s:  rdlen == 0 -- update?", fn));
+			}
 			rp_ex = NULL;
 			rdlen_ex = 0;
 		}
 
-		DX (syslog (LOG_DEBUG, "%s: add the extracted RR to matching RR_List.", fn));
+		if (T.debug > 3) {
+			DX (syslog (LOG_DEBUG, "%s: add the extracted RR to matching RR_List.", fn));
+		}
 		rrl_tmp = rr_list_add (rc_tmp->rrl, r_ttl, rdlen_ex, rp_ex);
 		if (!rrl_tmp) {
 			syslog (LOG_NOTICE, "%s: rr_list_add failed.", fn);
@@ -752,7 +888,9 @@ u_char *mesg_read_sec (int cnt, u_char * sec, u_char * begin, u_char * end, G_Li
 		rc_tmp->rrl = rrl_tmp;
 	}
 
-	DX (syslog (LOG_DEBUG, "%s: make each RRset from list.", fn));
+	if (T.debug > 3) {
+		DX (syslog (LOG_DEBUG, "%s: make each RRset from list.", fn));
+	}
 	rc_head->list_data = NULL;
 	for (gl_tmp = rc_head->next; gl_tmp->list_data; gl_tmp = gl_tmp->next) {
 		/* create complete RRset */
@@ -767,7 +905,9 @@ u_char *mesg_read_sec (int cnt, u_char * sec, u_char * begin, u_char * end, G_Li
 			goto error;
 		}
 		if (ls) {
-			DX (syslog (LOG_DEBUG, "%s: add it to R_List* ls.", fn));
+			if (T.debug > 3) {
+				DX (syslog (LOG_DEBUG, "%s: add it to R_List* ls.", fn));
+			}
 			if (list_add (ls, rrset_copy (rrs_tmp)) < 0) {
 				syslog (LOG_ERR, "%s: list_add failed.", fn);
 				goto error;
@@ -781,7 +921,9 @@ u_char *mesg_read_sec (int cnt, u_char * sec, u_char * begin, u_char * end, G_Li
 	/* free unused resources */
 	list_destroy (rc_head, rrset_couple_freev);
 
-	DX (syslog (LOG_DEBUG, "%s: end.", fn));
+	if (T.debug > 3) {
+		DX (syslog (LOG_DEBUG, "%s: end.", fn));
+	}
 	return rp;		/* <- next byte of this section */
 	/* NOTREACHED */
 
@@ -791,9 +933,9 @@ error:
 	return NULL;
 }
 
+/* XXX: unused? */
 u_char *mesg_read_sec_in_order (int cnt, u_char * sec, u_char * begin,
-				    u_char * end, G_List * ls)
-{
+				    u_char * end, G_List * ls) {
 
 	u_char *rp;		/* read pointer */
 	int i;
@@ -815,7 +957,9 @@ u_char *mesg_read_sec_in_order (int cnt, u_char * sec, u_char * begin,
 
 	RRset *rrs_new = NULL;	/* ALLOC'ED */
 
-	DX (syslog (LOG_DEBUG, "mesg_read_sec_in_order(): start\n"));
+	if (T.debug > 3) {
+		DX (syslog (LOG_DEBUG, "mesg_read_sec_in_order(): start\n"));
+	}
 
 	cnt = ntohs (cnt);
 
@@ -824,7 +968,8 @@ u_char *mesg_read_sec_in_order (int cnt, u_char * sec, u_char * begin,
 	for (i = 0; i < cnt; i++) {
 		/* parse header */
 		rname = rp;
-		rp = mesg_skip_dname (rp);
+		if ((rp = mesg_skip_dname (rp, end)) == NULL)
+			goto bad;
 		GETSHORT (r_type, rp);
 		GETSHORT (r_class, rp);
 		GETLONG (r_ttl, rp);
@@ -832,9 +977,13 @@ u_char *mesg_read_sec_in_order (int cnt, u_char * sec, u_char * begin,
 		rdp = rp;
 		rp += rdlen;
 
-		DX (syslog (LOG_DEBUG, "mesg_read_sec_in_order(): packet info: type = %s, class = %s, rd_length = %d\n", string_rtype (r_type), string_rclass (r_class), rdlen));
+		if (T.debug > 3) {
+			DX (syslog (LOG_DEBUG, "mesg_read_sec_in_order(): packet info: type = %s, class = %s, rd_length = %d\n", 
+					string_rtype (r_type), string_rclass (r_class), rdlen));
+		}
 
 		if (rp > end) {
+		  bad:
 			/* overrun!! */
 			syslog (LOG_WARNING, "mesg_read_sec_in_order(): message overrun\n");
 			goto error;
@@ -890,7 +1039,10 @@ u_char *mesg_read_sec_in_order (int cnt, u_char * sec, u_char * begin,
 	if (rrs_new) {
 		rrset_free (rrs_new);
 	}
-	DX (syslog (LOG_DEBUG, "mesg_read_sec_in_order(): finish.\n"));
+
+	if (T.debug > 3) {
+		DX (syslog (LOG_DEBUG, "mesg_read_sec_in_order(): finish.\n"));
+	}
 	return rp;		/* next byte of this section */
 	/* NOTREACHED */
 
@@ -903,20 +1055,23 @@ error:
 }
 
 
-/* not public function -- called by mesg_parse{,_upd} */
-static int mesg_parse_common (u_char * mesg, int mesg_len, G_List * an_list,
-		           G_List * ns_list, G_List * ar_list, int upd_flag)
-{
+int mesg_parse(u_char * mesg, int mesg_len, G_List * an_list,
+	       G_List * ns_list, G_List * ar_list) {
 
 	Mesg_Hdr *hdr;
 	u_char *rp;
 
+	if (mesg_len < sizeof(*hdr)) { /* for safety. */
+		syslog(LOG_NOTICE, "message overrun\n");
+		return(-1);
+	}
+
 	hdr = (Mesg_Hdr *) mesg;
 	rp = (u_char *) (hdr + 1);	/* read next of the header */
 
 	if (hdr->qdcnt) {
 		/* skip question section */
-		SKIP_DNAME (rp);
+		rp = mesg_skip_dname (rp, mesg + mesg_len);
 		rp += 4;	/* sizeof qtype and qclass */
 		if (rp > mesg +mesg_len) {
 			syslog (LOG_NOTICE, "message overrun\n");
@@ -940,15 +1095,3 @@ static int mesg_parse_common (u_char * mesg, int mesg_len, G_List * an_list,
 	}
 	return 0;
 }
-
-int mesg_parse (u_char * mesg, int mesg_len, G_List * an_list,
-		    G_List * ns_list, G_List * ar_list)
-{
-	return mesg_parse_common (mesg, mesg_len, an_list, ns_list, ar_list, 0);
-}
-
-int mesg_parse_upd (u_char * mesg, int mesg_len, G_List * an_list,
-		        G_List * ns_list, G_List * ar_list)
-{
-	return mesg_parse_common (mesg, mesg_len, an_list, ns_list, ar_list, 1);
-}
