@@ -63,7 +63,7 @@
  ****************************************************************************/
 
 /*
- * <$Id: read_config.c,v 3.20 2000/04/02 12:35:26 dillema Exp $>
+ * <$Id: read_config.c,v 3.23 2000/09/28 11:52:35 dillema Exp $>
  */
 
 #include "tot.h"
@@ -77,10 +77,9 @@ COPYRIGHT2(
 COPYRIGHT3("Written by: F. W. Dillema. feico@pasta.cs.uit.no.\n
 based on code by: Yusuke DOI, Keio Univ. Murai Lab.");
 
-CVSID("$Id: read_config.c,v 3.20 2000/04/02 12:35:26 dillema Exp $");
+CVSID("$Id: read_config.c,v 3.23 2000/09/28 11:52:35 dillema Exp $");
 
-int read_config (char *config_file)
-{
+int read_config (char *config_file) {
 	char buf[512];
 	char *cp;
 	struct sockaddr *sa_p = NULL;
@@ -118,7 +117,7 @@ int read_config (char *config_file)
 		if (!args[0]) /* empty line (or only comment) */
 			continue;
 
-		if (strcasecmp (args[0], "forwarder") == 0) {
+		if (!strcasecmp (args[0], "forwarder")) {
 			if (argcnt < 2 || argcnt > 4 ) {
 				syslog (LOG_ERR, "line %d: invalid format: forwarder <IPaddr> [port <port>]", linenum);
 				errcnt++;
@@ -127,7 +126,7 @@ int read_config (char *config_file)
 
 				i = 2;
 				if (argcnt > i) {
-					if (strcasecmp (args[i], "port") == 0) {
+					if (!strcasecmp (args[i], "port")) {
 						if (argcnt >= ++i) {
 							port = atoi (args[i++]);
 						}
@@ -144,7 +143,7 @@ int read_config (char *config_file)
 					}	
 				}
 				sa_p = net_parse_alloc_addr (args[1], port, &sa_p_len);
-				if (sa_p && fwd_add (sa_p, sa_p_len) == 0) {
+				if (sa_p && !fwd_add (sa_p, sa_p_len)) {
 					syslog (LOG_INFO, "forwarder %s added, port %d", args[1], port);
 				}
 				else {
@@ -152,7 +151,7 @@ int read_config (char *config_file)
 					errcnt++;
 				}
 			}
-		} else if (strcasecmp (args[0], "prefix") == 0) {
+		} else if (!strcasecmp (args[0], "prefix")) {
 			if (argcnt != 2) {
 				syslog (LOG_ERR, "line %d: invalid format: prefix <IPv6prefix>", linenum);
 				errcnt++;
@@ -164,7 +163,7 @@ int read_config (char *config_file)
 					syslog (LOG_INFO, "prefix %d added: %s", T.prefixnum, args[1]);
 				}
 			}
-		} else if (strcasecmp (args[0], "retry") == 0) {
+		} else if (!strcasecmp (args[0], "retry")) {
 			if (argcnt != 2) {
 				syslog (LOG_ERR, "line %d: invalid format: retry <seconds>", linenum);
 				errcnt++;
@@ -173,7 +172,7 @@ int read_config (char *config_file)
 			}
 		}
 #ifdef SCOPED_REWRITE
-		else if (strcasecmp(args[0], "scoped") == 0) {
+		else if (!strcasecmp(args[0], "scoped")) {
 			if (argcnt != 4) {
 				syslog (LOG_ERR, "line %d: invalid format: scoped <from> <to> <plen>", linenum);
 				errcnt++;
@@ -186,6 +185,16 @@ int read_config (char *config_file)
 				}
 			}
 		}
+#endif
+#ifdef STF
+		else if  (!strcasecmp(args[0], "stf")) {
+			if (argcnt != 1) {
+				syslog (LOG_ERR, "line %d: invalid format: stf", linenum);
+				errcnt++;
+			} else {
+				T.stf = 1;
+			}
+		}
 #endif
 		else {
 			syslog (LOG_WARNING, "line %d: unknown keyword in config file: %s",
